/***
*		wye.c -- midi controlled y-valve
*		 1992, 1994, 1997  Another Option
***/

/*
============================================================================
|    DATE    | WHAT
============================================================================
   24 Jan 92 | created
	31 Jan 92 | reroute only note ons and offs; keep controller messages in
	          | the same pipeline.  !!should be configurable
	 1 Feb 92 | it is configurable, and control passing fixed
   26 Jan 93 | massaged for sas/c v6.0
   14 Feb 93 | massaged for bplib.lib; use bplib hiloSwitch
	14 Nov 93 | added snapshot & sysex, new window title
	-----------------------------------------------------------------------
   23 Jan 94 | new ui; expanded reroute based on bypass.c
	 4 Dec 97 | added branched note array; cleaned up poop. AND you really ought to
	          | add: a sysex menu; pipe for controller; selecting multiple ccs to
				 | reroute
*/

#define DBUG 0

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <string.h>
#include <intuition/intuition.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <clib/alib_protos.h>
#include <clib/alib_stdio_protos.h>

#include "bars.h"
#include "bplib.h"
#include "barsmacros.h"
#include "proto/wye_protos.h"
#include "wyew3.h"
#include "snapshotgadget.h"
#include "toolid.h"

#include "/icons/WyeDown.h"
#include "/icons/WyeUp.h"

#if DBUG
 #define FILEDUMP 1
 #include "debug.h"
#endif

static UBYTE const versiontag[] = "$VER: Wye Valve 40.01 (04.12.97)";

struct YTool {		/* a Tool structure for the tool */
	struct Tool tool;
	UBYTE state;
	UBYTE controller;   /* cc# that changes tool state */
	UBYTE midi_route;   /* bits for which midi events to wye */
	BYTE rr_controller; /* cc# to reroute; -1 = all */
	UBYTE branched[128];/* count of notes that got branched */
};

/* state bits */
#define ST_WYE        1
#define DISCARD_SYSEX 2
#define DISCARD_CTRL  4
#define USE_SYSEX     8

/* mask for what is real */
#define VALID_STATE_BITS 0x07

/*** bits for midi_route ***/
#define MR_NOTE   1
#define MR_BEND   2
#define MR_MTOUCH 4
#define MR_PTOUCH 8
#define MR_PGM   16
#define MR_CCHG  32
#define MR_SYSEX 64

#define VALID_MIDI_BITS 0x7F

/*
#define GDOWN 100
#define GUP 200
*/

/* refresh which gadgets bits */
#define REFRESH_STATE 1
#define REFRESH_MIDI 2

#define TOOL_SIZE sizeof(struct YTool) /* for ToolMaster */

extern struct IntuitionBase *IntuitionBase;
extern struct GfxBase *GfxBase;
extern struct LayersBase *LayersBase;
extern struct DOSBase *DOSBase;
extern struct ExecBase *SysBase;
extern struct Functions *functions;

/*==========================================================================
* inittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: sets tool defaults. 
*=========================================================================*/

static void __regargs inittool(tool)

struct YTool *tool;

{
	tool->state = DISCARD_CTRL | DISCARD_SYSEX;
	tool->controller = 66;
	tool->rr_controller = -1;
	tool->midi_route = MR_NOTE;
	tool->tool.touched |= TOUCH_INIT;
}

#define tw tool->tool.window

/********** refresh tool edit window gadgets ***************/


static void __regargs drawmidi(struct YTool *tool)
{
	if (tw && RefreshOK(&tool->tool)) {
		functions->SelectEmbossed(tw,RR_NOTE,tool->midi_route & MR_NOTE);
		functions->SelectEmbossed(tw,RR_BEND,tool->midi_route & MR_BEND);
		functions->SelectEmbossed(tw,RR_PGM,tool->midi_route & MR_PGM);
		functions->SelectEmbossed(tw,RR_MTOUCH,tool->midi_route & MR_MTOUCH);
		functions->SelectEmbossed(tw,RR_PTOUCH,tool->midi_route & MR_PTOUCH);
		functions->SelectEmbossed(tw,RR_SYSEX,tool->midi_route & MR_SYSEX);
		functions->SelectEmbossed(tw,RR_CCHG,tool->midi_route & MR_CCHG);
		EnableFatProp(tw,RR_CONTROLLER,RR_CONTROLLER+GUP,RR_CONTROLLER+GDOWN,
		              tool->midi_route & MR_CCHG);
		if (tool->midi_route & MR_CCHG) {
			functions->ModifyEmbossedProp(tw,RR_CONTROLLER,tool->rr_controller,0,0,0,0,0);
			functions->DrawEmbossedProp(tw,RR_CONTROLLER);
		}
	}
}


static void __regargs drawstate(struct YTool *tool, UBYTE doall)

{
	if (tw && RefreshOK(&tool->tool)) {
		functions->SelectEmbossed(tw,WYE,tool->state & ST_WYE);
		functions->SelectEmbossed(tw,PASS,!(tool->state & ST_WYE));
		if (doall) {
			functions->ModifyEmbossedProp(tw,CONTROLLER,tool->controller-63,0,0,0,0,0);
			functions->DrawEmbossedProp(tw,CONTROLLER);
			functions->SelectEmbossed(tw,DIS_CONTROL,tool->state & DISCARD_CTRL);
			functions->SelectEmbossed(tw,DIS_SYSEX,tool->state & DISCARD_SYSEX);
		}
	}
}

#define SX_LENGTH 4

/*==========================================================================
* dosysex
*
*  INPUT: pointer to event
* RETURN: pointer to event
*
* EFFECT: gets new tool state from sysex message
*=========================================================================*/

static __inline struct Event * dosysex(struct YTool *tool,
                                       struct StringEvent *event)

{
	register unsigned char *sxdata;
	UBYTE refresh = 0; /* and also a flag to discard the event */

	if (sxdata = checksysex(event,TOOL_SYSEX_ID,tool->tool.toolid)) {
		if (*sxdata != MIDI_EOX) {
			tool->state = (*sxdata) & VALID_STATE_BITS;
			refresh = REFRESH_STATE;
			if (*(++sxdata) != MIDI_EOX) {
				if ((*sxdata > 62) && (*sxdata < 121)) tool->controller = *sxdata;
				if (*(++sxdata) != MIDI_EOX) {
					tool->midi_route = (*sxdata) & VALID_MIDI_BITS;
					refresh |= REFRESH_MIDI;
					if (*(++sxdata) != MIDI_EOX) {
						tool->rr_controller = *sxdata;
					}
				}
			}
		}
		if (refresh & REFRESH_STATE) drawstate(tool,1);
		if (refresh & REFRESH_MIDI) drawmidi(tool);

		if (tool->state & DISCARD_SYSEX) refresh=1; /* flag a delete */
		else refresh=0;
	}
	else { /* deal with alien sysex */
		if ((tool->midi_route & MR_SYSEX) && (tool->state & ST_WYE)) {
			if (tool->tool.branch) {
				event->tool=tool->tool.branch;
			}
			else refresh=1;
		}
	}

	if (refresh) {
		functions->freestring(event->string);
		functions->freeevent((struct Event *)event);
		event=NULL;
	}

	return((struct Event *)event);
}

/*==========================================================================
* processevent
*
*  INPUT: pointer to Event
* RETURN: pointer to Event
*
* EFFECT: first, check for a control change we respond to and deal with it.
*         branch messages if supposed to and can, or free them.
*         all other messages are passed down the pipeline.
*=========================================================================*/

static struct Event * __stdargs __saveds processevent(struct Event *event)

{
	struct YTool *tool = (struct YTool *)event->tool;
//	struct Event *off;
	UBYTE doit;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);
	event->tool = event->tool->next;


	if (IsSysEx(event)) {
		event = dosysex(tool,(struct StringEvent *)event);
	}
	else if (IsRealTime(event)) {
		if (IsCtrlChg(event) && tool->controller && (event->byte1==tool->controller)) {
			doit = tool->state;
			if (SwitchOn(event->byte2)) tool->state |= ST_WYE;
			else if (SwitchOff(event->byte2)) tool->state &= ~ST_WYE;
			if (tool->state & DISCARD_CTRL) {
				/* make sure we are not supposed to reroute this cc */
			   if ( !((tool->midi_route & MR_CCHG) && 
				      ((event->byte1 == tool->rr_controller) || 
					    (tool->rr_controller == 121)) ) ) {
					functions->freeevent(event);
					event=NULL;
				}
			}
			/* redraw only if state was changed */
			if (doit != tool->state) { 
				drawstate(tool,0);
			}
		}
		if (IsNoteOff(event)) {
			if (tool->branched[event->byte1]) {
				tool->branched[event->byte1]--;
				if (tool->tool.branch) {
					event->tool = tool->tool.branch;
					event->type |= EVENT_BRANCH;
				}
				else {
					functions->freeevent(event);
					event = NULL;
				}
			}
		}
		else if (event && IsChannelMsg(event)) {
			if (tool->state & ST_WYE) {
				if (IsNoteOn(event) && (tool->midi_route & MR_NOTE)) doit=1;
				else if (IsPitchBend(event) && (tool->midi_route & MR_BEND)) doit=1;
				else if (IsMonoTouch(event) && (tool->midi_route & MR_MTOUCH)) doit=1;
				else if (IsPolyTouch(event) && (tool->midi_route & MR_PTOUCH)) doit=1;
				else if (IsPgmChg(event) && (tool->midi_route & MR_PGM)) doit=1;
				else if (IsCtrlChg(event) && (tool->midi_route & MR_CCHG)) {
					if (event->byte1==tool->rr_controller) doit=1;
				}
				else doit=0;

				if (doit) {
					if (tool->tool.branch) {
						event->tool = tool->tool.branch;
						event->type |= EVENT_BRANCH;
						if (IsNoteOn(event)) {
							tool->branched[event->byte1]++;
/*							if (off = functions->dupeevent(event)) {
								event->next = off;
								off->status = MIDI_NOTEOFF;
								off->time += ((struct NoteEvent *)event)->duration;
							}
							else {
								functions->freeevent(event);
								event=NULL;
							}
*/
						}
					}
					else {
						functions->freeevent(event);
						event = NULL;
					}
				}
			}
		} /* if ChannelMsg() */
	} /* if RealTime() */

	return(event);
}


/*==========================================================================
* dosnapshot
*
*  INPUT: pointer to YTool
* RETURN: none
*
* EFFECT: makes a sysex message and posts it
*=========================================================================*/

static void __inline dosnapshot(struct YTool *tool)

{
	UBYTE sx[SX_LENGTH];
	struct StringEvent *event;
	
	memcpy((char *)sx,(char *)(&tool->state),SX_LENGTH);
	if (event = BuildSnapshot(&tool->tool, 
	                          functions->running ? functions->timenow : -1, sx,
									  SX_LENGTH)) {
		PostSnapshot(&tool->tool,event);
	}
}


/* for display of a prop gadget value */
#pragma msg 84 ignore push
#define XOffset 45 /* relative to g->LeftEdge + g->Width */
#pragma msg 84 ignore pop
#define YOffset 6  /* relative to g->TopEdge */

/* pens for pot value text */
#define TEXTCOLOR 1
#define BACKCOLOR 0

/* display pot value offset by 63; 0==off */

static long __stdargs __saveds hiloSwOff(struct Window *w, struct Gadget *g, long h)

{
	char text[4];
	
	Move(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset);
	SetAPen(w->RPort,TEXTCOLOR);
	SetBPen(w->RPort,BACKCOLOR);
	SetDrMd(w->RPort,JAM2);
	h += 63; /* offset: minimum switch controller = 64; allow for off==0 */
	if (h > 63) {
		sprintf(text,"%3ld",h);
	}
	else sprintf(text,"%s","Off");
	Text(w->RPort,text,3);
	return(h);
}

/* display pot value where 121 == All */

static long __stdargs __saveds hiloAll(struct Window *w, struct Gadget *g, long h)

{
	char text[4];
	
	Move(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset);
	SetAPen(w->RPort,TEXTCOLOR);
	SetBPen(w->RPort,BACKCOLOR);
	SetDrMd(w->RPort,JAM2);
	if (h < 121) {
		sprintf(text,"%3ld",h);
	}
	else sprintf(text,"%s","All");
	Text(w->RPort,text,3);
	return(h);
}


/*==========================================================================
* edittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: ui handler
*=========================================================================*/

static void __stdargs __saveds edittool(tool)

struct YTool *tool;

{
	struct IntuiMessage *message;
	long class; 
	long code;
	struct Gadget *gadget;
	struct NewWindow *nw;
	UBYTE refresh;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);

	nw=OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,20);
	if (!nw) return;
	AddSnapshotGadget(tw);
	PrintIText(tw->RPort,&IntuiTextList1,0,0);
	EmbossOnGList(tw,&GadgetList1,GCOUNT,1);
	functions->FatEmbossedPropOn(tw,CONTROLLER,CONTROLLER+GDOWN,CONTROLLER+GUP,
	                                hiloSwOff,58,1);
	functions->FatEmbossedPropOn(tw,RR_CONTROLLER,RR_CONTROLLER+GDOWN,RR_CONTROLLER+GUP,
	                             hiloAll,122,1);
	SetMenuStrip(tw,&MenuList1);
	refresh=255;
	SetRefreshOK(&tool->tool);

	/*** ui handling ***/
	for ( ; ; ) {
		if (refresh & REFRESH_STATE) drawstate(tool,1);
		if (refresh & REFRESH_MIDI) drawmidi(tool);
		refresh=0;
		message = (struct IntuiMessage *)functions->GetIntuiMessage(tw);
		class = message->Class;
		code = message->Code;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW) {
			ResetRefreshOK(&tool->tool);
			break;
		}
		else if (class == GADGETUP) {
			if (gadget->GadgetID == SNAPSHOT_GADGET_ID) dosnapshot(tool);
			else switch (gadget->GadgetID) {
				case PASS :
					tool->state &= ~ST_WYE;
					refresh=REFRESH_STATE;
					break;
				case WYE :
					tool->state |= ST_WYE;
					refresh=REFRESH_STATE;
					break;
				case DIS_CONTROL :
					tool->state ^= DISCARD_CTRL;
					refresh=REFRESH_STATE;
					break;
				case DIS_SYSEX :
					tool->state ^= DISCARD_SYSEX;
					refresh=REFRESH_STATE;
					break;
			/* midi gadgets */
				case RR_ALLMIDI :
					tool->midi_route = VALID_MIDI_BITS;
					refresh=REFRESH_MIDI;
					break;
				case RR_NOTE :
					tool->midi_route ^= MR_NOTE;
					refresh=REFRESH_MIDI;
					break;
				case RR_BEND :
					tool->midi_route ^= MR_BEND;
					refresh=REFRESH_MIDI;
					break;
				case RR_PGM :
					tool->midi_route ^= MR_PGM;
					refresh=REFRESH_MIDI;
					break;
				case RR_CCHG :
					tool->midi_route ^= MR_CCHG;
					refresh=REFRESH_MIDI;
					break;
				case RR_MTOUCH :
					tool->midi_route ^= MR_MTOUCH;
					refresh=REFRESH_MIDI;
					break;
				case RR_PTOUCH :
					tool->midi_route ^= MR_PTOUCH;
					refresh=REFRESH_MIDI;
					break;
				case RR_SYSEX :
					tool->midi_route ^= MR_SYSEX;
					refresh=REFRESH_MIDI;
					break;
			}
		}
		else if (class == GADGETDOWN) {
			switch (gadget->GadgetID) {
				case CONTROLLER :
					tool->controller=(*functions->DragEmbossedProp)(tw,CONTROLLER);
					refresh=REFRESH_STATE;
					break;
				case CONTROLLER+GDOWN :
					tool->controller=(*functions->ShiftEmbossedProp)(tw,CONTROLLER,-1,0);
					refresh=REFRESH_STATE;
					break;
				case CONTROLLER+GUP :
					tool->controller=(*functions->ShiftEmbossedProp)(tw,CONTROLLER,1,0);
					refresh=REFRESH_STATE;
					break;
				case RR_CONTROLLER :
					tool->rr_controller=(*functions->DragEmbossedProp)(tw,RR_CONTROLLER);
					refresh=REFRESH_MIDI;
					break;
				case RR_CONTROLLER+GDOWN :
					tool->rr_controller=(*functions->ShiftEmbossedProp)(tw,RR_CONTROLLER,-1,0);
					refresh=REFRESH_MIDI;
					break;
				case RR_CONTROLLER+GUP :
					tool->rr_controller=(*functions->ShiftEmbossedProp)(tw,RR_CONTROLLER,1,0);
					refresh=REFRESH_MIDI;
					break;
			}
		}
	}
	EmbossOffGList(tw,&GadgetList1,GCOUNT);
	functions->FatEmbossedPropOff(tw,CONTROLLER,CONTROLLER+GDOWN,CONTROLLER+GUP);
	functions->FatEmbossedPropOff(tw,RR_CONTROLLER,RR_CONTROLLER+GDOWN,RR_CONTROLLER+GUP);
	NewCloseEmbossedWindow(&tool->tool,nw);
}

/*==========================================================================
* loadtool
*
*  INPUT: file id, size of chunk to read
* RETURN: pointer to allocated tool, or NULL
*
* EFFECT: allocates and reads tool data from a file. resets TOUCH_REFRESHOK
*         in tool.touched
*=========================================================================*/

static struct Tool * __stdargs __saveds loadtool(long file, long size)

{
	struct YTool *tool;
	
	if (tool = (struct YTool *)functions->myalloc(TOOL_SIZE,0)) {
		memset((char *)tool,0,TOOL_SIZE);
		functions->fastread(file,(char *)tool,size);
		tool->tool.touched &= ~TOUCH_REFRESHOK;
		size -= TOOL_SIZE;
	}
	if (size > 0) { /* leftovers */
		functions->fastseek(file,size,0);
/*
		if (tool) {
			functions->myfree((char *)tool,TOOL_SIZE);
			tool=NULL;
		}
*/
	}
	return((struct Tool *)tool);
}
	

/********************************************/
static void __stdargs __saveds transportcode(command, time, end)

register long command;
register long time;
register long end;

{
	register struct YTool *tool;
	register struct Track *track = (struct Track *)functions->tracklist;

	if (command == TC_STOP) {
//		if (time < 0) time = 0;
		for ( ; track ; track=track->next ) {
			tool = (struct YTool *)track->toollist;
			for ( ; tool ; tool=(struct YTool *)tool->tool.next ) {
				if (tool->tool.toolid == WYYE_TOOLID) memset((char *)tool->branched,0,128);
			}
		}
	}
}

/*******************************************/
static void __stdargs __saveds removetool(void)
{
	functions->removetransport((void *)transportcode);
}


static struct ToolMaster master;

/*==========================================================================
* inittoolmaster
*
*  INPUT: none
* RETURN: pointer to tool master struct
*
* EFFECT: initializes.
*=========================================================================*/

struct ToolMaster * __stdargs __saveds inittoolmaster()

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = TOOL_SIZE;
	master.toolid = WYYE_TOOLID;
	master.edittool = edittool;
	master.processevent = processevent;
	master.loadtool = loadtool;
	master.image = &YIconDown;
	master.upimage = &YIconUp;
	strcpy(master.name,WTITLE);
	master.tooltype = TOOL_NORMAL|TOOL_BRANCHOUT;

	functions->installtransportpl((void *)transportcode,0);

	return(&master);
}
